<?php

namespace Neklo\Scheduler\Observer;

use Magento\Framework\Console\CLI;
use Neklo\Scheduler\Api\Data\JobInterface;
use Neklo\Scheduler\Model\ResourceModel\Job\CollectionFactory;
use Magento\Cron\Model\Schedule;

/**
 * Class Observer
 * @package Neklo\Scheduler\Observer
 */
class Observer extends \Magento\Cron\Observer\ProcessCronQueueObserver
{
    /**
     * @var CollectionFactory
     */
    protected $jobCollectionFactory;
    /**
     * @var \Neklo\Scheduler\Model\JobFactory
     */
    protected $jobFactory;
    /**
     * @var \Psr\Log\LoggerInterface
     */
    protected $logger;

    /**
     * Observer constructor.
     * @param \Magento\Framework\ObjectManagerInterface $objectManager
     * @param \Magento\Cron\Model\ScheduleFactory $scheduleFactory
     * @param \Magento\Framework\App\CacheInterface $cache
     * @param \Magento\Cron\Model\ConfigInterface $config
     * @param \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
     * @param \Magento\Framework\App\Console\Request $request
     * @param \Magento\Framework\ShellInterface $shell
     * @param \Magento\Framework\Stdlib\DateTime\DateTime $dateTime
     * @param \Magento\Framework\Process\PhpExecutableFinderFactory $phpExecutableFinderFactory
     * @param \Psr\Log\LoggerInterface $logger
     * @param \Magento\Framework\App\State $state
     * @param \Magento\Framework\Lock\LockManagerInterface $lockManager
     * @param CollectionFactory $jobCollectionFactory
     * @param \Neklo\Scheduler\Model\JobFactory $jobFactory
     */
    public function __construct(
        \Magento\Framework\ObjectManagerInterface $objectManager,
        \Magento\Cron\Model\ScheduleFactory $scheduleFactory,
        \Magento\Framework\App\CacheInterface $cache,
        \Magento\Cron\Model\ConfigInterface $config,
        \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig,
        \Magento\Framework\App\Console\Request $request,
        \Magento\Framework\ShellInterface $shell,
        \Magento\Framework\Stdlib\DateTime\DateTime $dateTime,
        \Magento\Framework\Process\PhpExecutableFinderFactory $phpExecutableFinderFactory,
        \Psr\Log\LoggerInterface $logger,
        \Magento\Framework\App\State $state,
        \Magento\Framework\Lock\LockManagerInterface $lockManager,
        CollectionFactory $jobCollectionFactory,
        \Neklo\Scheduler\Model\JobFactory $jobFactory
    )
    {
        parent::__construct($objectManager, $scheduleFactory, $cache, $config,
            $scopeConfig, $request, $shell, $dateTime, $phpExecutableFinderFactory, $logger, $state, $lockManager);
        $this->jobCollectionFactory = $jobCollectionFactory;
        $this->jobFactory = $jobFactory;
        $this->logger = $logger;
    }

    /**
     * @param \Magento\Framework\Event\Observer $observer
     */
    public function execute(\Magento\Framework\Event\Observer $observer)
    {
        $pendingJobs = $this->_getPendingSchedules();
        $currentTime = $this->dateTime->gmtTimestamp();
        $reqGroup = $this->_request->getParam('group');
        /* @var $jobs \Neklo\Scheduler\Model\ResourceModel\Job\Collection */
        $jobs = $this->jobCollectionFactory->create();
        // $jobs->addFieldToFilter(JobInterface::JOB_ENABLED, 1);
        $jobGroupsRoot = $this->groupJobs($jobs);

        foreach ($jobGroupsRoot as $groupId => $jobsRoot) {
            if ($reqGroup !== null
                && $reqGroup !== '\'' . ($groupId) . '\''
                && $reqGroup !== $groupId) {
                continue;
            }
            if ($this->_request->getParam(self::STANDALONE_PROCESS_STARTED) !== '1' &&
                $this->_scopeConfig->getValue('system/cron/' . $groupId . '/use_separate_process', \Magento\Store\Model\ScopeInterface::SCOPE_STORE) == 1
            ) {
                $this->_shell->execute(
                    'php %s cron:run --group=' . $groupId
                    . ' --' . CLI::INPUT_KEY_BOOTSTRAP . '='
                    . self::STANDALONE_PROCESS_STARTED . '=1',
                    [
                        BP . '/bin/magento'
                    ]
                );
                continue;
            }

            // Note: we acquire the lock here instead of above, as it should be taken by standalone (child) process,
            // not by the parent process.
            if (!$this->lockManager->lock(self::LOCK_PREFIX . $groupId, self::LOCK_TIMEOUT)) {

                $message = sprintf("Could not acquire lock for cron group: %s, skipping run", $groupId);

                if ($this->logger != null) {
                    $this->logger->warning($message);
                } else {
                    echo $message . PHP_EOL;
                }
                continue;
            }

            $this->_cleanup($groupId);
            $this->_generate($groupId);

            foreach ($pendingJobs as $schedule) {
                $jobConfig = isset($jobsRoot[$schedule->getJobCode()])
                    ? $jobsRoot[$schedule->getJobCode()] : null;
                if (!$jobConfig) {
                    continue;
                } else

                    if (!$jobConfig->isEnabled()) {
                        $schedule->setStatus(Schedule::STATUS_ERROR)
                            ->setMessages('Run disabled by job config')->save();
                        continue;
                    }

                $scheduledTime = strtotime($schedule->getScheduledAt());
                if ($scheduledTime > $currentTime) {
                    continue;
                }

                try {
                    if ($schedule->tryLockJob()) {
                        $this->_runJob($scheduledTime, $currentTime,
                            $jobConfig->getData(), $schedule, $groupId);
                    }
                } catch (\Exception $e) {
                    $schedule->setMessages($e->getMessage());
                }
                $schedule->save();
            }

            $this->lockManager->unlock(self::LOCK_PREFIX . $groupId);
        }

        $this->updateJobs();
    }

    /**
     *
     * @param \Neklo\Scheduler\Model\ResourceModel\Job\Collection $jobs
     * @return array
     */
    protected function groupJobs(
        \Neklo\Scheduler\Model\ResourceModel\Job\Collection $jobs
    )
    {
        $ret = [];
        foreach ($jobs->getItems() as $job) {
            if (!isset($ret[$job->getGroup()])) {
                $ret[$job->getGroup()] = [];
            }
            $ret[$job->getGroup()][$job->getName()] = $job;
        }
        return $ret;
    }

    /**
     *
     */
    protected function updateJobs()
    {
        $jobGroups = $this->_config->getJobs();
        foreach ($jobGroups as $group => $jobs) {
            foreach ($jobs as $jobName => $job) {
                $model = $this->jobFactory->create();
                $model = $model->load($jobName, JobInterface::JOB_NAME);
                if (!$model->getId()) {
                    $model->setName($jobName)
                        ->setMethod($job[JobInterface::JOB_METHOD])
                        ->setInstance($job[JobInterface::JOB_CLASS])
                        ->setEnabled(true)
                        ->setGroup($group);
                    if (isset($job[JobInterface::JOB_SCHEDULE])) {
                        $model->setSchedule($job[JobInterface::JOB_SCHEDULE]);
                    }
                    if (isset($job[JobInterface::JOB_CONFIG])) {
                        $model->setConfigPath($job[JobInterface::JOB_CONFIG]);
                    }
                    $model->save();
                }
            }
        }
    }
}
